# --------------------------------------------------------------------------------------------------
# Copyright 2008, Yi Ren@BlockM

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
#
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

# This code is designed to fulfill part of my Master's research at ODE|UM.
# --------------------------------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Version History:
#   2008.Jun.10     v1.0    Initial release
#   2008.Jul.04     v1.02   Continuity C1
#   2008.Jul.23     v1.03   Continuity C2, Constraints and variation on Component 1, 2, 3
#   2008.Jul.26     v1.04   Continuity C2, Constraints and variation on Component 4
#   2008.Aug.14     v1.05   Continuity C2, Constraints and variation on Component 5


#-----------------------------------------------------------------------------

require  'sketchup.rb' # Link Sketchup 6.0
require  File.dirname(__FILE__) + '/Bezier_curve.rb' # Bezier curve class
require  File.dirname(__FILE__) + '/Bezier_surf.rb' # Bezier surface class
require  File.dirname(__FILE__) + '/Mid_point.rb' # Find mid_point for Bezier curves
require  File.dirname(__FILE__) + '/Rational.rb' # Rational number
require  File.dirname(__FILE__) + '/Wireframe.rb' # Wireframe methods


class Matchbox

@@pluginDirectory = File.dirname(__FILE__)
@@title = "Run Matchbox"

##
# Main entry point
##
    def Matchbox::init()

      # Somehow we do not have a current directory. Alert the user and end.
      if !File.directory?(@@pluginDirectory)
        alert(@@title + " plugin has encountered an error:\n" +
              "Could not find the plugin directory. " +
              "Please re-install the plugin.")
        return
      end

      # make sure the plugin was installed correctly
      tempImgDirectory = @@pluginDirectory + "/images"
      if !File.directory?(tempImgDirectory) || 
         !File.writable?(tempImgDirectory) 
        alert(@@title + " plugin has encountered an error:\n" +
              "Could not find or could not write temporary images to:\n" +
              "[" + tempImgDirectory + "]\n" + 
              "Please re-install the plugin.")
        return
      end

      cmd = UI::Command.new(@@title) {main()}
      cmd.small_icon = tempImgDirectory + "/MatchboxButtontb.png"
      cmd.large_icon = tempImgDirectory + "/MatchboxButtontb.png"
      cmd.status_bar_text = cmd.tooltip = "Create a random vehicle"
      cmd.menu_text = @@title
      matchboxToolbar = UI::Toolbar.new(@@title) 
      matchboxToolbar.add_item(cmd)
      matchboxToolbar.show

      UI::menu('Tools').add_item(cmd);
    end 

    def Matchbox::main()
#===================================================================== Global =================================================================================
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
            # Global Constants
            @dimension_coef = 0.15
            @eps = 0.000001
            @origin = Geom::Point3d.new(0, 0, 0)
            
            ex = Geom::Vector3d.new(1, 0, 0)
            ey = Geom::Vector3d.new(0, 1, 0)
            ez = Geom::Vector3d.new(0, 0, 1)
            
            # Global Variables
            #@center = Geom::Point3d.new((@box_l+@box_r)/2, 0, 0)
            
            # Global Parameters
            @numseg = 8

            model = Sketchup.active_model
            aCar = model.active_entities.add_group
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
#===================================================================== Global =================================================================================

#===================================================================== Modeling ===============================================================================
    # Component 1: Front window, roof and rear window
    # 17 variables
            c1 = aCar.entities.add_group
            profile1 = Hash.new
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
            # Local Parameters
            @eta = 0.8 # eta is a coef to smooth the curve, see notes on constraint on component 1
            @well_scale = 1.5
            @r_wheel = 77.0
            @roof_end_pt_x = 800.0 # higher limit on @roof_end_pt
            @bw_end_pt_x = 900.0 # higher limit on @bw_end_pt
            @roof_end_pt_z = @origin.z + @r_wheel*@well_scale + 50.0 # lower limit on @roof_end_pt
            @bw_end_pt_z = @roof_end_pt_z - 20.0 # lower limit on @bw_end_pt
            @hood_u_n = 100 # to avoid surface intrution (@nose_mid_pt is on the negative side of center plane)
            
            # Variables and Curves with Constraints
            @hood_end_pt = Geom::Point3d.new(-20.0 + 100*rand(),0.0,220.0 + 30*rand())
            @roof_start_pt = Geom::Point3d.new(@hood_end_pt.x + 240.0 + 10*rand(),0.0,@hood_end_pt.z + 90.0 + 30*rand())
            
            temp = (@roof_start_pt - @hood_end_pt).normalize.x * (0.8 + 0.2 * rand())
            @fw_vec = Geom::Vector3d.new(temp, 0.0, Math.sqrt(1.0-temp*temp))
            
            # We restrict @fw_n so that there will be no front window that bend down (m<0)
            @fw_n = (@roof_start_pt.z - @hood_end_pt.z)*0.8/ @fw_vec.z +  (@roof_start_pt.z - @hood_end_pt.z)*0.2/ @fw_vec.z*rand()
            
            @fw_v_n = 90.0  
            
            temp = Wireframe.link_tan(@hood_end_pt, @roof_start_pt, @fw_vec, @fw_n, @numseg)
            profile1['fw'] = temp['all_pt']
            profile1['fw_ctrl_pt'] = temp['mid_pt']
            
            # The position of @roof_end_pt is constrained by the requirement that roof_ctrl_pt shall be "between" @roof_end_pt and @roof_start_pt
            # See notes on "Constraints on Component 1, 2"
            # Here we first randomize @roof_end_pt.x and then find the acceptable domain for @roof_end_pt.z
            ux = @hood_end_pt.x - profile1['fw_ctrl_pt'].x
            uz = @hood_end_pt.z - profile1['fw_ctrl_pt'].z
            vx = @roof_start_pt.x - profile1['fw_ctrl_pt'].x
            vz = @roof_start_pt.z - profile1['fw_ctrl_pt'].z
            d = vx*uz-vz*ux
            xb = @roof_end_pt_x - @roof_start_pt.x
            xbb = xb*0.9
            zb = @hood_end_pt.z + 100 - @roof_start_pt.z
            zbb = @roof_end_pt_z - @roof_start_pt.z
            
            if ux>0&&uz<0 #case 1
                if vz < 0
                    UI.messagebox('Error generating shape_Error#_c1_10: Shape not acceptable.')
                else
                    eta = xb/vx*0.5
                    m = [(xbb*uz-zbb*ux)/d, zbb/vz].max #lower bound for psi
                    n = [(xb*uz-zb*ux)/d, xb/vx].min
                    if m > n
                        UI.messagebox('Error generating shape_Error#_c1_20: Boundary condition not acceptable')
                    else
                        psi = m + (n - m)*rand()
                        mm = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, 0, vx/ux*(eta-psi)].max
                        nn = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux].min
                        if mm > nn
                            UI.messagebox('Error generating shape_Error#_c1_01: Wrong range for phi')
                        end
                    end
                end
                flag = 1
            elsif ux<0&&uz<0 #case 2
                if vz<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx, zbb/vz].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, (vx*uz*eta-zb*ux)/d, (zb-uz*eta*eta)/vz, eta].max
                    n = [(xb*uz-zbb*ux)/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                elsif vz>0&&d<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, eta, (vx*uz*eta-zb*ux)/d].max
                    n = [(xb*uz-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz].min
                else
                    eta = [(xbb*uz-zb*ux)/d, xb/vx].min*0.5
                    m = [(xb*uz-zbb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz, (vx*uz*eta-zb*ux)/d].min
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_21: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(zb-psi*vz)/uz, (xb-psi*vx)/ux, 0].max
                    nn = [(zbb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_02: Wrong range for phi')
                    end
                end
                flag = 2
            elsif ux<0&&uz>0 #case 3
                if vz >0
                    eta = [xb/vx, (xb*uz-zb*ux)/d, zb/vz].min*0.5
                    m = [(xbb*uz-zbb*ux)/d, xbb/vx, eta, (vx*uz*eta-zbb*ux)/d, (zbb-uz*eta*eta)/vz].max
                    n = [(xb*uz-zb*ux)/d, zb/vz, (xb-ux*eta*eta)/vx].min
                else
                    eta = [xb/vx, zbb/vz, (xbb*uz-zbb*ux)/d].min*0.5
                    m = [(xb*uz-zb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zbb*ux)/d, (vx*uz*eta-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zbb-uz*eta*eta)/vz].min
                end
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_22: Boundary condition not acceptable')
                else
                    psi = m +(n - m)*rand()
                    mm = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux, 0].max
                    nn = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_03: Wrong range for phi')
                    end
                end
                
                flag = 3
            elsif ux<0 && uz == 0 #case 4
                
                if vz<0
                    eta = [-zbb*ux/d, xb/vx, zbb/vz].min*0.5
                    m = [-zb*ux/d, xbb/vx, -zb*ux/d, zb/vz, eta].max
                    n = [-zbb*ux/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                else
                    eta = [zb/vz, xb/vx,].min*0.5
                    m = [zbb/vz, xbb/vx, eta].max
                    n = [zb/vz, (xb-ux*eta*eta)/vx].min
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_23: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(xb-psi*vx)/ux, 0.0].max
                    nn = [(xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_04: Wrong range for phi')
                    end
                end
                
                flag = 4
            else
                UI.messagebox('Error generating shape_Error#_c1_13: Shape not acceptable.')
            end
            
            phi = mm + (nn - mm)*rand()
            @roof_end_pt = Geom::Point3d.new(@roof_start_pt.x + phi*ux + psi*vx, 0, @roof_start_pt.z + phi*uz + psi*vz)
            temp = Wireframe.link_cur(@hood_end_pt, profile1['fw_ctrl_pt'], @roof_start_pt, @roof_end_pt, 1, @numseg)
            profile1['roof'] = temp['all_pt']
            profile1['roof_ctrl_pt'] = temp['mid_pt']
            
            ux = @roof_start_pt.x - profile1['roof_ctrl_pt'].x
            uz = @roof_start_pt.z - profile1['roof_ctrl_pt'].z
            vx = @roof_end_pt.x - profile1['roof_ctrl_pt'].x
            vz = @roof_end_pt.z - profile1['roof_ctrl_pt'].z
            d = vx*uz-vz*ux
            xb = @bw_end_pt_x - @roof_end_pt.x
            xbb = xb*0.1
            zb = 1000
            zbb = @bw_end_pt_z - @roof_end_pt.z
            
            if ux>0&&uz<0 #case 1
                if vz < 0
                    UI.messagebox('Error generating shape_Error#_c1_40: Shape not acceptable.')
                else
                    eta = xb/vx*0.5
                    m = [(xbb*uz-zbb*ux)/d, zbb/vz].max #lower bound for psi
                    n = [(xb*uz-zb*ux)/d, xb/vx].min
                    if m > n
                        UI.messagebox('Error generating shape_Error#_c1_50: Boundary condition not acceptable')
                    else
                        psi = m + (n - m)*rand()
                        mm = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, 0, vx/ux*(eta-psi)].max
                        nn = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux].min
                        if mm > nn
                            UI.messagebox('Error generating shape_Error#_c1_31: Wrong range for phi')
                        end
                    end
                end
                flag = 1
            elsif ux<0&&uz<0 #case 2
                
                if vz<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx, zbb/vz].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, (vx*uz*eta-zb*ux)/d, (zb-uz*eta*eta)/vz, eta].max
                    n = [(xb*uz-zbb*ux)/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                elsif vz>0&&d<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, eta, (vx*uz*eta-zb*ux)/d].max
                    n = [(xb*uz-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz].min
                else
                    eta = [(xbb*uz-zb*ux)/d, xb/vx].min*0.5
                    m = [(xb*uz-zbb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz, (vx*uz*eta-zb*ux)/d].min
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_51: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(zb-psi*vz)/uz, (xb-psi*vx)/ux, 0].max
                    nn = [(zbb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_32: Wrong range for phi')
                    end
                end
                flag = 2
            elsif ux<0&&uz>0 #case 3
                
                if vz >0
                    eta = [xb/vx, (xb*uz-zb*ux)/d, zb/vz].min*0.5
                    m = [(xbb*uz-zbb*ux)/d, xbb/vx, eta, (vx*uz*eta-zbb*ux)/d, (zbb-uz*eta*eta)/vz].max
                    n = [(xb*uz-zb*ux)/d, zb/vz, (xb-ux*eta*eta)/vx].min
                else
                    eta = [xb/vx, zbb/vz, (xbb*uz-zbb*ux)/d].min*0.5
                    m = [(xb*uz-zb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zbb*ux)/d, (vx*uz*eta-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zbb-uz*eta*eta)/vz].min
                end
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_52: Boundary condition not acceptable')
                else
                    psi = m +(n - m)*rand()
                    mm = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux, 0].max
                    nn = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_33: Wrong range for phi')
                    end
                end
                
                flag = 3
            elsif ux<0 && uz == 0 #case 4
                
                if vz<0
                    eta = [-zbb*ux/d, xb/vx, zbb/vz].min*0.5
                    m = [-zb*ux/d, xbb/vx, -zb*ux/d, zb/vz, eta].max
                    n = [-zbb*ux/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                else
                    eta = [zb/vz, xb/vx,].min*0.5
                    m = [zbb/vz, xbb/vx, eta].max
                    n = [zb/vz, (xb-ux*eta*eta)/vx].min
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c1_53: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(xb-psi*vx)/ux, 0.0].max
                    nn = [(xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c1_34: Wrong range for phi')
                    end
                end
                
                flag = 4
            else
                UI.messagebox('Error generating shape_Error#_c1_43: Shape not acceptable.')
            end
            phi = mm + (nn - mm)*rand()
            @bw_end_pt = Geom::Point3d.new(@roof_end_pt.x + phi*ux + psi*vx, 0, @roof_end_pt.z + phi*uz + psi*vz)
            temp = Wireframe.link_cur(@roof_start_pt, profile1['roof_ctrl_pt'], @roof_end_pt, @bw_end_pt, 1, @numseg)
            profile1['bw'] = temp['all_pt']
            profile1['bw_ctrl_pt'] = temp['mid_pt']
            
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
            @c1_t = 0.8 + 0.2*rand() #scale-copy the profile on xz plane to the side curve
            @hood_u_pt_x = (@bw_end_pt.x - @hood_end_pt.x)*(1-@c1_t)*(0.5 + 0.5*rand()) #move side curve of component1 on the x axis
            @hood_u_pt_z = -5*rand() #move side curve of component1 on the z axis
            @hood_u_pt = Geom::Point3d.new(@hood_end_pt.x + @hood_u_pt_x, 180+20*rand(), @hood_end_pt.z + @hood_u_pt_z)
            @fw_u_pt = Geom::Point3d.new(@hood_end_pt.x + (@roof_start_pt.x - @hood_end_pt.x)*@c1_t + @hood_u_pt_x, @hood_u_pt.y, @hood_end_pt.z + (@roof_start_pt.z - @hood_end_pt.z)*@c1_t + @hood_u_pt_z)
            @roof_pt = Geom::Point3d.new(@hood_end_pt.x + (@roof_end_pt.x - @hood_end_pt.x)*@c1_t + @hood_u_pt_x, @hood_u_pt.y, @hood_end_pt.z + (@roof_end_pt.z - @hood_end_pt.z)*@c1_t + @hood_u_pt_z)
            @bw_pt = Geom::Point3d.new(@hood_end_pt.x + (@bw_end_pt.x - @hood_end_pt.x)*@c1_t + @hood_u_pt_x, @hood_u_pt.y, @hood_end_pt.z + (@bw_end_pt.z - @hood_end_pt.z)*@c1_t + @hood_u_pt_z)
            
            @fw_v_n = @fw_n*@c1_t
            
            c1_r1 = Geom::Transformation.new(@hood_u_pt, @fw_vec, -0.2*rand()) #Component1 rotation1
            [@fw_u_pt, @roof_pt, @bw_pt].each {|pt| pt = pt.transform!(c1_r1)}
            fw_v_hash = Wireframe.link_tan(@hood_u_pt, @fw_u_pt, @fw_vec, @fw_v_n, @numseg)
            
            c1_r2 = Geom::Transformation.new(@fw_u_pt, @fw_u_pt - fw_v_hash['mid_pt'] , 0.0*rand()) #Component1 rotation1
            [@roof_pt, @bw_pt].each {|pt| pt = pt.transform!(c1_r2)}
            roof_v_hash = Wireframe.link_cur(@hood_u_pt, fw_v_hash['mid_pt'], @fw_u_pt, @roof_pt, 1, @numseg)
            
            c1_r3 = Geom::Transformation.new(@roof_pt, @roof_pt - roof_v_hash['mid_pt'], 0) #Component1 rotation1
            @bw_pt = @bw_pt.transform!(c1_r3)
            bw_v_hash = Wireframe.link_cur(@fw_u_pt, roof_v_hash['mid_pt'], @roof_pt, @bw_pt, 1, @numseg)
            
            # Surface
            hood_u_hash = Wireframe.link_tan(@hood_end_pt, @hood_u_pt, ey, @hood_u_n, @numseg)
            
            # To reduce the number of variables, we assume that the mid pts of hood_center, roof and hood_u curves have 
            # the same propotion to their total y length.
            k = hood_u_hash['mid_pt'].y / @hood_u_pt.y
            
            fw_center_n =  k * fw_v_hash['mid_pt'].y
            fw_center = Mid_point.surftan1(profile1['fw_ctrl_pt'], ey, fw_center_n)
            
            fw_u_n = k * @fw_u_pt.y
            fw_u_hash = Wireframe.link_tan(@roof_start_pt, @fw_u_pt, ey, fw_u_n, @numseg)
            
            pq = Mid_point.curvpqxz(@hood_end_pt, profile1['fw_ctrl_pt'], @roof_start_pt, @roof_end_pt)
            p = pq[0]
            q = pq[1]
            
            a = @roof_start_pt - profile1['fw_ctrl_pt'] + ey.transform(Geom::Transformation.new(fw_u_n - fw_center_n))
            b = @hood_end_pt - profile1['fw_ctrl_pt'] + ey.transform(Geom::Transformation.new(@hood_u_n - fw_center_n))
            
            mid_pt = Geom::Point3d.new
            mid_pt.x = a.x*p + b.x*q + fw_u_hash['mid_pt'].x
            mid_pt.y = a.y*p + b.y*q + fw_u_hash['mid_pt'].y
            mid_pt.z = a.z*p + b.z*q + fw_u_hash['mid_pt'].z
            #bc = Bezier_curve.new
            roof_u_hash = Hash.new
            roof_u_hash['mid_pt'] = mid_pt
            #roof_u_hash['all_pt'] = bc.curve([@roof_end_pt, mid_pt, @roof_pt], @numseg, nil)
            
            roof_center = Geom::Point3d.new
            roof_center.x = fw_u_hash['mid_pt'].x + a.x*Math.sqrt(q.abs)
            roof_center.y = fw_u_hash['mid_pt'].y + a.y*Math.sqrt(q.abs)
            roof_center.z = fw_u_hash['mid_pt'].z + a.z*Math.sqrt(q.abs)

            roof_center_n = roof_center.y
            roof_u_n = roof_u_hash['mid_pt'].y
            pq = Mid_point.curvpqxz(@roof_start_pt, profile1['roof_ctrl_pt'], @roof_end_pt, @bw_end_pt)
            p = pq[0]
            q = pq[1]
            
            a = @roof_end_pt - profile1['roof_ctrl_pt'] + ey.transform(Geom::Transformation.new(roof_u_n - roof_center_n))
            b = @roof_start_pt - profile1['roof_ctrl_pt'] + ey.transform(Geom::Transformation.new(fw_u_n - roof_center_n))
            
            mid_pt = Geom::Point3d.new
            mid_pt.x = a.x*p + b.x*q + roof_u_hash['mid_pt'].x
            mid_pt.y = a.y*p + b.y*q + roof_u_hash['mid_pt'].y
            mid_pt.z = a.z*p + b.z*q + roof_u_hash['mid_pt'].z
            bw_u_hash = Hash.new
            bw_u_hash['mid_pt'] = mid_pt
            
            bw_center = Geom::Point3d.new
            bw_center.x = roof_u_hash['mid_pt'].x + a.x*Math.sqrt(q.abs)
            bw_center.y = roof_u_hash['mid_pt'].y + a.y*Math.sqrt(q.abs)
            bw_center.z = roof_u_hash['mid_pt'].z + a.z*Math.sqrt(q.abs)
            
            data = Hash.new
            data['uOrder'] = 2
            data['vOrder'] = 2
            data['uSteps'] = @numseg
            data['vSteps'] = @numseg
            
            data['ctrlPts'] = [@hood_end_pt, hood_u_hash['mid_pt'], @hood_u_pt, profile1['fw_ctrl_pt'], fw_center, fw_v_hash['mid_pt'], @roof_start_pt, fw_u_hash['mid_pt'], @fw_u_pt]
            fw_surf = Bezier_surf.new(data)
            fw_surf.create_patch(c1.entities)
            
            data['ctrlPts'] = [@roof_start_pt, fw_u_hash['mid_pt'], @fw_u_pt, profile1['roof_ctrl_pt'], roof_center, roof_v_hash['mid_pt'], @roof_end_pt, roof_u_hash['mid_pt'], @roof_pt]
            roof_surf = Bezier_surf.new(data)
            roof_surf.create_patch(c1.entities)
            
            data['ctrlPts'] = [@roof_end_pt, roof_u_hash['mid_pt'], @roof_pt, profile1['bw_ctrl_pt'], bw_center, bw_v_hash['mid_pt'], @bw_end_pt, bw_u_hash['mid_pt'], @bw_pt]
            bw_surf = Bezier_surf.new(data)
            bw_surf.create_patch(c1.entities)

#--------------------------------------------------------------------------------------------------------------------------------------------------------------
    # Component 2: Nose and Hood
    # 10 variables
            c2 = aCar.entities.add_group
            profile2 = Hash.new
            
            # Parameters
            @box_w = 240 #car width = wheel.y
            @nose_mid_pt = Geom::Point3d.new(-250, 0, 0)
            
            # Variables
            @hood_start_pt = Geom::Point3d.new(@nose_mid_pt.x  + 50*rand() ,0,120 + 40*rand())
            temp = (@hood_start_pt - @hood_end_pt).normalize.z * 0.8 + (@hood_start_pt - @hood_end_pt).normalize.z * 0.2 * rand()
            @hood_vec = Geom::Vector3d.new(-Math.sqrt(1.0-temp*temp), 0.0, temp)
            @hood_n = (@hood_start_pt.x - @hood_end_pt.x)*0.8/ @hood_vec.x +  (@hood_start_pt.x - @hood_end_pt.x) * 0.2 /  @hood_vec.x*rand()
            
            
            temp = Wireframe.link_tan(@hood_end_pt, @hood_start_pt, @hood_vec, @hood_n, @numseg)
            profile2['hood'] = temp['all_pt']
            profile2['hood_ctrl_pt'] = temp['mid_pt']
            
            temp = Wireframe.link_cur(@hood_end_pt, profile2['hood_ctrl_pt'], @hood_start_pt, @nose_mid_pt, 1, @numseg)
            profile2['nose'] = temp['all_pt']
            profile2['nose_ctrl_pt'] = temp['mid_pt']
            
            @nose_u_pt = Geom::Point3d.new(@hood_start_pt.x + 5 + 50*rand(), @hood_u_pt.y - 10 - 60*rand(), @hood_start_pt.z)
            
            temp = (@nose_u_pt - @hood_u_pt).normalize.z * 0.5 + (@nose_u_pt - @hood_u_pt).normalize.z * 0.5 * rand()
            hood_v_vec = Geom::Vector3d.new((@nose_u_pt - @hood_u_pt).normalize.x, (@nose_u_pt - @hood_u_pt).normalize.y, temp)
            hood_v_vec = hood_v_vec.normalize
            @hood_v_n = (@nose_u_pt.x - @hood_u_pt.x)*0.8/ hood_v_vec.x +  (@nose_u_pt.x - @hood_u_pt.x)*0.2/ hood_v_vec.x*rand()
            
            hood_v_hash = Wireframe.link_tan(@hood_u_pt, @nose_u_pt, hood_v_vec, @hood_v_n, @numseg)
            
            temp = hood_u_hash['mid_pt'].y/((hood_u_hash['mid_pt'] - @nose_mid_pt).length)
            hood_svec_x = -temp+temp*rand()
            @hood_svec = Geom::Vector3d.new(hood_svec_x, Math.sqrt(1-hood_svec_x*hood_svec_x))
            hood_center = Geom.intersect_line_plane([profile2['hood_ctrl_pt'], Geom::Vector3d.new(0,1,0)], [hood_u_hash['mid_pt'], @hood_svec])
            mid_pt1 = Geom.intersect_line_plane([@hood_start_pt, Geom::Vector3d.new(0,1,0)], [hood_u_hash['mid_pt'], @hood_svec])
            mid_pt2 = Geom.intersect_line_plane([@nose_mid_pt, Geom::Vector3d.new(0,1,0)], [hood_u_hash['mid_pt'], @hood_svec])
            
            nose_u_hash = Hash.new
            nose_u_hash['mid_pt'] = mid_pt1
            
            @head_end_pt = Geom::Point3d.new(mid_pt2.x + 10 + 20*rand(), @nose_u_pt.y - (@nose_u_pt.y - mid_pt2.y)/2*rand())
            head_u_hash = Hash.new
            head_u_hash['mid_pt'] = mid_pt2
            
            nose_v_hash = Wireframe.link_cur(@hood_u_pt, hood_v_hash['mid_pt'], @nose_u_pt, @head_end_pt, 1, @numseg)
            
            pq = Mid_point.curvpqxz(hood_u_hash['mid_pt'], hood_center, nose_u_hash['mid_pt'], head_u_hash['mid_pt'])
            q = pq[1]
            a = nose_u_hash['mid_pt'] - hood_center
            nose_center = Geom::Point3d.new
            nose_center.x = nose_u_hash['mid_pt'].x + a.x*Math.sqrt(q.abs)
            nose_center.y = nose_u_hash['mid_pt'].y + a.y*Math.sqrt(q.abs)
            nose_center.z = nose_u_hash['mid_pt'].z + a.z*Math.sqrt(q.abs)
            
            data['ctrlPts'] = [@nose_mid_pt, head_u_hash['mid_pt'], @head_end_pt, profile2['nose_ctrl_pt'], nose_center, nose_v_hash['mid_pt'], @hood_start_pt, nose_u_hash['mid_pt'], @nose_u_pt]
            nose_surf = Bezier_surf.new(data)
            nose_surf.create_patch(c2.entities)
            
            data['ctrlPts'] = [@hood_start_pt, nose_u_hash['mid_pt'], @nose_u_pt, profile2['hood_ctrl_pt'], hood_center, hood_v_hash['mid_pt'], @hood_end_pt, hood_u_hash['mid_pt'], @hood_u_pt]
            hood_surf = Bezier_surf.new(data)
            hood_surf.create_patch(c2.entities)
            
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
    # Component 3: Wheel well
    # 9 variables
            c3 = aCar.entities.add_group
            
            # Variables
            @head_v_n = 80.0 # Parameter?
            @belt_cur_n = 20*rand()
            
            # head_u1_pt
            @head_u1_pt_x = @hood_u_pt.x - 2.0*@r_wheel*@well_scale
            #@head_u1_pt_x = 100
            @head_u1_pt_y = @box_w - 0.0*@well_scale
            ux = @nose_mid_pt.y - head_u_hash['mid_pt'].y
            uz = - @nose_mid_pt.x + head_u_hash['mid_pt'].x
            vx = @head_end_pt.y - head_u_hash['mid_pt'].y
            vz = - @head_end_pt.x + head_u_hash['mid_pt'].x
            d = vx*uz-vz*ux
            xb = @head_u1_pt_y - @head_end_pt.y
            xbb = (@head_u1_pt_y - @head_end_pt.y)*0.9
            zb = 0.0
            zbb = (xbb/vx*vz)*1.1
            
            if ux<0 #case 2
                if vz<0
                    eta = [-zbb*ux/d, xb/vx, zbb/vz].min*0.5
                    m = [-zb*ux/d, xbb/vx, -zb*ux/d, zb/vz, eta].max
                    n = [-zbb*ux/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                else
                    UI.messagebox('Error generating shape_Error#_c2_10: Shape not acceptable.')
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c2_20: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(xb-psi*vx)/ux, 0.0].max
                    nn = [(xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c2_01: Wrong range for phi')
                    end
                end
                flag = 2
            else
                UI.messagebox('Error generating shape_Error#_c2_11: Shape not acceptable.')
            end
            phi = mm + (nn - mm)*rand()
            @head_u1_pt = Geom::Point3d.new(@head_end_pt.x - psi*vz, @head_end_pt.y + phi*ux + psi*vx, 0)
            head_u1_hash = Wireframe.link_cur(@nose_mid_pt, head_u_hash['mid_pt'], @head_end_pt, @head_u1_pt, 2, @numseg)
            
            # Center line
            @head_u2_vec = (@nose_u_pt - nose_u_hash['mid_pt']).normalize
            
            @well_vec = Geom::Vector3d.new(0.0, 1.0, -1.0*rand()).normalize
            temp_t = Geom::Transformation.new(0.5)
            head_center_vec = @head_u2_vec.transform!(temp_t) + (@head_end_pt - head_u_hash['mid_pt']).normalize.transform!(temp_t)
            
            temp1 = (@hood_u_pt.y - head_u1_hash['mid_pt'].y)/Math.sqrt((@hood_u_pt.y - head_u1_hash['mid_pt'].y)*(@hood_u_pt.y - head_u1_hash['mid_pt'].y)+(@hood_u_pt.x - head_u1_hash['mid_pt'].x)*(@hood_u_pt.x - head_u1_hash['mid_pt'].x)) # lower limit for well center plane
            temp = @head_u1_pt.y - 10.0
            temp2 = (temp - head_u1_hash['mid_pt'].y)/Math.sqrt((temp - head_u1_hash['mid_pt'].y)*(temp - head_u1_hash['mid_pt'].y)+(@hood_u_pt.x - head_u1_hash['mid_pt'].x)*(@hood_u_pt.x - head_u1_hash['mid_pt'].x)) # upper limit for well center plane
            
            temp = [temp1, temp2].min + 1.0*([temp1, temp2].max-[temp1, temp2].min)*rand()
            well_v_vec = Geom::Vector3d.new(temp, -Math.sqrt(1-temp*temp), 0.0)
            mid_pt1 = Geom.intersect_line_plane([@nose_u_pt, @head_u2_vec], [head_u1_hash['mid_pt'], well_v_vec]) #mid_pt of head_u2_hash
            mid_pt2 = Geom.intersect_line_plane([@hood_u_pt, @well_vec], [head_u1_hash['mid_pt'], well_v_vec]) #mid_pt of well_u_hash
            head_center = Geom.intersect_line_plane([nose_v_hash['mid_pt'], head_center_vec], [head_u1_hash['mid_pt'], well_v_vec]) #mid_pt of well_u_hash
            # Center line
            
            # Parameters
            @fwheel_x = @head_u1_pt.x + @well_scale*@r_wheel #wheel coordinates
            @fwheel_z = @nose_mid_pt.z
            
            pq = Mid_point.curvpqxy(@nose_mid_pt, head_u_hash['mid_pt'], @head_end_pt, @head_u1_pt)
            p = pq[0]
            q = pq[1]
            
            @head_u2_pt = Geom::Point3d.new(@fwheel_x - @r_wheel*@well_scale/2.0, @head_u1_pt.y, @r_wheel*@well_scale + @fwheel_z) #30deg
            head_v_hash = Wireframe.link_tan(@head_u2_pt, @head_u1_pt, Geom::Vector3d.new(-Math.sqrt(3.0)/2.0, 0, -1.0/2.0), @head_v_n, @numseg)
            
            #bc = Bezier_curve.new
            head_u2_hash = Hash.new
            head_u2_hash['mid_pt'] = mid_pt1
            
            # @well_pt
            ux = @head_u1_pt.x - head_v_hash['mid_pt'].x
            uz = @head_u1_pt.z - head_v_hash['mid_pt'].z
            vx = @head_u2_pt.x - head_v_hash['mid_pt'].x
            vz = @head_u2_pt.z - head_v_hash['mid_pt'].z
            d = vx*uz-vz*ux
            xb = (mid_pt2.x - @head_u2_pt.x)*1.2
            xbb = mid_pt2.x - @head_u2_pt.x
            zb = (mid_pt2.z - @head_u2_pt.z) # Constrained by well_mid_pt
            zbb = [xb/vx*vz, zb].min*0.5
            
            if ux>0&&uz<0 #case 1
                if vz < 0
                    UI.messagebox('Error generating shape_Error#_c3_30: Shape not acceptable.')
                else
                    eta = xb/vx*0.5
                    m = [(xbb*uz-zbb*ux)/d, zbb/vz, (xbb-eta*eta*ux)/vx, (vx*uz*eta-zbb*ux)/d].max #lower bound for psi
                    n = [(xb*uz-zb*ux)/d, xb/vx, (zb-eta*eta*uz)/vz].min
                    
                    if m > n
                        UI.messagebox('Error generating shape_Error#_c3_40: Boundary condition not acceptable')
                    else
                        psi = m + (n - m)*rand()
                        mm = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, 0.0, vx/ux*(eta-psi)].max
                        nn = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux, eta*eta].min
                        if mm > nn
                            UI.messagebox('Error generating shape_Error#_c3_21: Wrong range for phi')
                        end
                    end
                end
                flag = 1
            elsif ux<0&&uz<0 #case 2
                
                if vz<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx, zbb/vz].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, (vx*uz*eta-zb*ux)/d, (zb-uz*eta*eta)/vz, eta].max
                    n = [(xb*uz-zbb*ux)/d, zbb/vz, (xb-ux*eta*eta)/vx].min
                elsif vz>0&&d<0
                    eta = [(xb*uz-zbb*ux)/d, xb/vx].min*0.5
                    m = [(xbb*uz-zb*ux)/d, xbb/vx, eta, (vx*uz*eta-zb*ux)/d].max
                    n = [(xb*uz-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz].min
                else
                    eta = [(xbb*uz-zb*ux)/d, xb/vx].min*0.5
                    m = [(xb*uz-zbb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zb*ux)/d, (xb-ux*eta*eta)/vx, (zb-uz*eta*eta)/vz, (vx*uz*eta-zb*ux)/d].min
                end
                
                if m > n
                    UI.messagebox('Error generating shape_Error#_c3_41: Boundary condition not acceptable')
                else
                    psi = m + (n - m)*rand()
                    mm = [(zb-psi*vz)/uz, (xb-psi*vx)/ux, 0.0].max
                    nn = [(zbb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c3_22: Wrong range for phi')
                    end
                end
                flag = 2
            elsif ux<0&&uz>0 #case 3
                
                if vz >0
                    eta = [xb/vx, (xb*uz-zb*ux)/d, zb/vz].min*0.5
                    m = [(xbb*uz-zbb*ux)/d, xbb/vx, eta, (vx*uz*eta-zbb*ux)/d, (zbb-uz*eta*eta)/vz].max
                    n = [(xb*uz-zb*ux)/d, zb/vz, (xb-ux*eta*eta)/vx].min
                else
                    eta = [xb/vx, zbb/vz, (xbb*uz-zbb*ux)/d].min*0.5
                    m = [(xb*uz-zb*ux)/d, xbb/vx, eta].max
                    n = [(xbb*uz-zbb*ux)/d, (vx*uz*eta-zbb*ux)/d, (xb-ux*eta*eta)/vx, (zbb-uz*eta*eta)/vz].min
                end
                if m > n
                    UI.messagebox('Error generating shape_Error#_c3_42: Boundary condition not acceptable')
                else
                    psi = m +(n - m)*rand()
                    mm = [(zbb-psi*vz)/uz, (xb-psi*vx)/ux, 0.0].max
                    nn = [(zb-psi*vz)/uz, (xbb-psi*vx)/ux, vx/ux*(eta-psi), eta*eta].min
                    if mm > nn
                        UI.messagebox('Error generating shape_Error#_c3_23: Wrong range for phi')
                    end
                end
                
                flag = 3
            else
                UI.messagebox('Error generating shape_Error#_c3_32: Shape not acceptable.')
            end
            phi = mm + (nn - mm)*rand()
            @well_pt = Geom::Point3d.new(@head_u2_pt.x + phi*ux + psi*vx, mid_pt2.y + 10, @head_u2_pt.z + phi*uz + psi*vz)
            well_v_hash = Wireframe.link_cur(@head_u1_pt, head_v_hash['mid_pt'], @head_u2_pt, @well_pt, 1, @numseg) #temporary
            
            #bc = Bezier_curve.new
            well_u_hash = Hash.new
            well_u_hash['mid_pt'] = mid_pt2
            
            pq = Mid_point.curvpqxz(head_u1_hash['mid_pt'], head_center, head_u2_hash['mid_pt'], well_u_hash['mid_pt'])
            q = pq[1]
            a = head_u2_hash['mid_pt'] - head_center
            
            well_center = Geom::Point3d.new
            well_center.x = head_u2_hash['mid_pt'].x + a.x*Math.sqrt(q.abs)
            well_center.y = head_u2_hash['mid_pt'].y + a.y*Math.sqrt(q.abs)
            well_center.z = head_u2_hash['mid_pt'].z + a.z*Math.sqrt(q.abs)
            
            @front_well_pt1 = Geom::Point3d.new(@fwheel_x + @r_wheel*@well_scale, @head_u1_pt.y, 0.0)
            
            front_well_u1_hash = Wireframe.link_cur(@head_u1_pt, head_v_hash['mid_pt'], @head_u2_pt, @front_well_pt1, 1, @numseg)
            @fwell_vec = (@well_pt - well_u_hash['mid_pt']).normalize
            @fwell_vec = Geom::Vector3d.new(@fwell_vec.x, @fwell_vec.y-0.0*rand(), @fwell_vec.z)
            @fwell_vec = @fwell_vec.normalize
            
            @front_well_pt2 = Geom::Point3d.new(@well_pt.x - well_v_hash['mid_pt'].x + @front_well_pt1.x, @front_well_pt1.y - 10*rand(), @front_well_pt1.z)
            front_well_u2_hash = Wireframe.link_tan(@well_pt, @front_well_pt2, @fwell_vec, @belt_cur_n, @numseg)
            front_well_center = Geom.intersect_line_plane([well_v_hash['mid_pt'], well_v_hash['mid_pt']-well_center], [@well_pt, Geom::Vector3d.new(0.0, 1.0, 0.0)])
            front_well_v_hash = Wireframe.link_tan(@front_well_pt2, @front_well_pt1, Geom::Vector3d.new(-1.0, 0.0, 0.0), 0.9*(@front_well_pt2.x - @front_well_pt1.x), @numseg)
            
            # Front wheel extension
            fw1 = Geom::Point3d.new(@fwheel_x - @r_wheel*1.2, @box_w, 0.0)
            fw2 = Geom::Point3d.new(@fwheel_x, @box_w, @r_wheel*1.2)
            fw3 = Geom::Point3d.new(@fwheel_x + @r_wheel*1.2, @box_w, 0.0)
            fw12 = Wireframe.link_tan(fw2, fw1, Geom::Vector3d.new(-1, 0, 0), 70, @numseg)
            fw23 = Wireframe.link_cur(fw1, fw12['mid_pt'], fw2, fw3, 1, @numseg)
            
            #fw_mid2 = Geom.intersect_line_plane([@head_u2_pt, @head_u2_pt - nose_u_hash['mid_pt']], [Geom::Point3d.new(0.0, (@box_w*0.2+@head_u1_pt.y*0.8), 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            fw_mid2 = Geom.intersect_line_plane([@head_u2_pt, fw2 - @head_u2_pt], [Geom::Point3d.new(0.0, (@box_w*0.2+@head_u1_pt.y*0.8), 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            fw_mid1 = Geom.intersect_line_plane([@head_u1_pt, @head_u1_pt - head_u_hash['mid_pt']], [Geom::Point3d.new(0.0, (@box_w*0.2+@head_u1_pt.y*0.8), 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            #mid3 = Geom.intersect_line_plane([@front_well_pt1, @front_well_pt1 - front_well_v_hash['mid_pt']], [Geom::Point3d.new(0, (@box_w+@head_u1_pt.y)*0.5, 0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            fw_mid3 = @front_well_pt1 + (fw_mid1-@head_u1_pt).transform!(Geom::Transformation.scaling(-1, 1, 1))
            
            fw_center1 = Geom.intersect_line_plane([head_v_hash['mid_pt'], head_v_hash['mid_pt']-head_center], [Geom::Point3d.new(0.0, (@box_w+@head_u1_pt.y)*0.5, 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            fw_center2_hash = Wireframe.link_cur(fw_mid1, fw_center1, fw_mid2, fw_mid3, 1, @numseg)
            
            data['ctrlPts'] = [@head_end_pt, head_u1_hash['mid_pt'], @head_u1_pt, nose_v_hash['mid_pt'], head_center, head_v_hash['mid_pt'], @nose_u_pt, head_u2_hash['mid_pt'], @head_u2_pt]
            head_surf = Bezier_surf.new(data)
            head_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@nose_u_pt, head_u2_hash['mid_pt'], @head_u2_pt, hood_v_hash['mid_pt'], well_center, well_v_hash['mid_pt'], @hood_u_pt, well_u_hash['mid_pt'], @well_pt]
            well_surf = Bezier_surf.new(data)
            well_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@head_u2_pt, front_well_u1_hash['mid_pt'], @front_well_pt1, well_v_hash['mid_pt'], front_well_center, front_well_v_hash['mid_pt'], @well_pt, front_well_u2_hash['mid_pt'], @front_well_pt2]
            front_well_surf = Bezier_surf.new(data)
            front_well_surf.create_patch(c3.entities)
            
            c3_wheel = aCar.entities.add_group
            data['ctrlPts'] = [@head_u1_pt, fw_mid1, fw1, head_v_hash['mid_pt'], fw_center1, fw12['mid_pt'], @head_u2_pt, fw_mid2, fw2]
            front_well_ext1_surf = Bezier_surf.new(data)
            front_well_ext1_surf.create_patch(c3_wheel.entities)
            
            data['ctrlPts'] = [@head_u2_pt, fw_mid2, fw2, front_well_u1_hash['mid_pt'], fw_center2_hash['mid_pt'], fw23['mid_pt'], @front_well_pt1, fw_mid3, fw3]
            front_well_ext2_surf = Bezier_surf.new(data)
            front_well_ext2_surf.create_patch(c3_wheel.entities)
            
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
    # Component 4: Beltline
    # 2 variables
            c4 = aCar.entities.add_group
            
            @box_r = @bw_end_pt.x + 100
            @wheelbase = 10.0*@r_wheel
            @clearance = 0.7*@r_wheel
            @wheel_scale = 1
            @box_l = @origin.x-2*@r_wheel
            @box_u = 4.5*@r_wheel
            @rear_well_pt = Geom::Point3d.new(@head_u2_pt.x+@r_wheel*(0.5+0.5*@wheel_scale)*@well_scale+@wheelbase, @front_well_pt1.y, @r_wheel*@well_scale*@wheel_scale)
            @rear_well_pt2 = Geom::Point3d.new(@head_u2_pt.x+@r_wheel*(0.5-1.0*@wheel_scale)*@well_scale+@wheelbase, @front_well_pt1.y, 0)
            @rear_pt = Geom::Point3d.new(@rear_well_pt2.x + @r_wheel*@well_scale*@wheel_scale*2, @head_u1_pt.y, 0)
            @rear_well_u2_n = 120
            rear_well_u2_hash = Wireframe.link_tan(@rear_well_pt, @rear_well_pt2, Geom::Vector3d.new(-Math.sqrt(3.0)/2.0, 0, 1.0/2.0), @rear_well_u2_n, @numseg)
            rear_v_hash = Wireframe.link_cur(@rear_well_pt2, rear_well_u2_hash['mid_pt'], @rear_well_pt, @rear_pt, 1, @numseg)
            
            temp_vec = (@well_pt - well_v_hash['mid_pt']).normalize
            temp_pt = Geom.intersect_line_line([Geom::Point3d.new(@well_pt.x, 0, @well_pt.z), Geom::Vector3d.new(temp_vec.x, 0, temp_vec.z)], [Geom::Point3d.new(@rear_well_pt.x, 0, @rear_well_pt.z), Geom::Vector3d.new( - (Math.sqrt(3.0)/2.0), 0, 1.0/2.0)])
            temp2 = temp_pt.x/temp_vec.x # constraint on @rear_well_pt tangent
            belt_line_vec = Geom::Vector3d.new(temp_vec.x, (@box_w-@well_pt.y)/temp2*rand(), temp_vec.z)
            
            if belt_line_vec.z > 0
                temp1 = [(@rear_well_pt2.x - @well_pt.x)/belt_line_vec.x, (@box_w - @well_pt.y)/belt_line_vec.y, (@fw_u_pt.z - @well_pt.z)/belt_line_vec.z].min
            else
                temp1 = [(@rear_well_pt2.x - @well_pt.x)/belt_line_vec.x, (@box_w - @well_pt.y)/belt_line_vec.y, (@well_pt.z - @rear_well_pt2.z)/belt_line_vec.z].min
            end
            
            @belt_n = [temp1, temp2].min + ([temp1, temp2].max - [temp1, temp2].min)*rand()
            
            side_temp = Wireframe.link_tan(@well_pt, @rear_well_pt, belt_line_vec, @belt_n, @numseg*3)
            @side_low_pt = side_temp['all_pt'][@numseg]
            @side_high_pt = side_temp['all_pt'][@numseg*2]
            
            temp_vec1 = Geom::Vector3d.new()
            temp_vec1.x = @rear_well_pt.x*(2.0/3.0) - @well_pt.x*(4.0/3.0) + side_temp['mid_pt'].x*(2.0/3.0) # Tangent at center of side_temp, see notes
            temp_vec1.y = @rear_well_pt.y*(2.0/3.0) - @well_pt.y*(4.0/3.0) + side_temp['mid_pt'].y*(2.0/3.0)
            temp_vec1.z = @rear_well_pt.z*(2.0/3.0) - @well_pt.z*(4.0/3.0) + side_temp['mid_pt'].z*(2.0/3.0)
            temp_vec2 = Geom::Vector3d.new()
            temp_vec2.x = @rear_well_pt.x*(4.0/3.0) - @well_pt.x*(2.0/3.0) - side_temp['mid_pt'].x*(2.0/3.0)
            temp_vec2.y = @rear_well_pt.y*(4.0/3.0) - @well_pt.y*(2.0/3.0) - side_temp['mid_pt'].y*(2.0/3.0)
            temp_vec2.z = @rear_well_pt.z*(4.0/3.0) - @well_pt.z*(2.0/3.0) - side_temp['mid_pt'].z*(2.0/3.0)
            
            mid_pt1 = Geom.intersect_line_line([@well_pt, belt_line_vec], [@side_low_pt, temp_vec1])
            mid_pt2 = Geom.intersect_line_line([@side_high_pt, temp_vec2], [@side_low_pt, temp_vec1])
            mid_pt3 = Geom.intersect_line_line([@rear_well_pt, @rear_well_pt - side_temp['mid_pt']], [@side_high_pt, temp_vec2])
            
            side_low_v_hash = Hash.new
            side_low_v_hash['mid_pt'] = mid_pt1
            temp1 = (@bw_pt.y - well_u_hash['mid_pt'].y)/Math.sqrt((@bw_pt.y - well_u_hash['mid_pt'].y)*(@bw_pt.y - well_u_hash['mid_pt'].y)+(@bw_pt.x - well_u_hash['mid_pt'].x)*(@bw_pt.x - well_u_hash['mid_pt'].x)) # lower limit for well center plane
            temp = @rear_well_pt.y
            temp2 = (temp - well_u_hash['mid_pt'].y)/Math.sqrt((temp - well_u_hash['mid_pt'].y)*(temp - well_u_hash['mid_pt'].y)+(@rear_well_pt.x - well_u_hash['mid_pt'].x)*(@rear_well_pt.x - well_u_hash['mid_pt'].x)) # upper limit for well center plane
            temp = [temp1 + temp2].max
            
            belt_v_vec = Geom::Vector3d.new(temp, -Math.sqrt(1-temp*temp), 0.0) # belt_v_vec_temp
            mid_pt = Geom.intersect_line_plane([@fw_u_pt, (@fw_u_pt - fw_u_hash['mid_pt'])],[well_u_hash['mid_pt'], belt_v_vec]) # side_low_u mid_pt
            side_low_u_hash = Hash.new
            side_low_u_hash['mid_pt'] = mid_pt
            
            side_high_v_hash = Hash.new
            side_high_v_hash['mid_pt'] = mid_pt2
            side_high_center = Geom.intersect_line_plane([roof_v_hash['mid_pt'], roof_v_hash['mid_pt']-roof_center],[side_high_v_hash['mid_pt'], (side_high_v_hash['mid_pt']-@side_low_pt)*(@side_low_pt-side_low_u_hash['mid_pt'])])
            belt_v_vec = (side_high_center - side_low_u_hash['mid_pt'])*(side_low_u_hash['mid_pt'] - well_u_hash['mid_pt'])
            
            side_low_center = Geom.intersect_line_plane([fw_v_hash['mid_pt'], fw_v_hash['mid_pt']-fw_center],[well_u_hash['mid_pt'], belt_v_vec])
            
            temp_vec = ((side_low_v_hash['mid_pt'] - side_low_center)).normalize
            temp_vec_n = (side_low_v_hash['mid_pt'].z - front_well_u2_hash['mid_pt'].z)/temp_vec.z
            rocker_center = side_low_v_hash['mid_pt'] + temp_vec.transform!(Geom::Transformation.new(-temp_vec_n))
            
            temp_vec = ((@side_low_pt - side_low_u_hash['mid_pt'])).normalize
            temp_pt = Geom.intersect_line_line([@side_low_pt, (@side_low_pt-side_low_u_hash['mid_pt'])], [side_high_v_hash['mid_pt'], (side_high_v_hash['mid_pt']-side_high_center)])
            
            temp_vec1 = (side_high_v_hash['mid_pt'] - rocker_center).normalize
            temp_vec2 = (@side_low_pt - side_low_u_hash['mid_pt']).normalize
            ux = temp_vec1.x
            uz = temp_vec1.z
            vx = temp_vec2.x
            vz = temp_vec2.z
            ax = @side_low_pt.x + temp_vec2.x - (rocker_center.x + temp_vec1.x)
            az = @side_low_pt.z + temp_vec2.z - (rocker_center.z + temp_vec1.z)
            phi = (ax*vz-az*vx)/(ux*vz-uz*vx)
            temp_pt2 = rocker_center + temp_vec1.transform!(Geom::Transformation.new(phi))
            temp = (temp_pt2.z - @side_low_pt.z)/temp_vec.z*1.1
            
            if (temp_pt.z - @side_low_pt.z)/temp_vec.z > 0
                if (temp_pt.z - @side_low_pt.z)/temp_vec.z*0.5 > temp
                    temp_vec_n = [temp, @belt_cur_n].max
                else
                    UI.messagebox('Error generating shape_Error#_c4_80: Side Error')
                end
            else
                temp_vec_n = [temp, @belt_cur_n].max
            end
            mid_pt = @side_low_pt + temp_vec.transform!(Geom::Transformation.new(temp_vec_n)) # rocker_u mid_pt
            
            rocker_u_hash = Hash.new
            rocker_u_hash['mid_pt'] = mid_pt
            
            @rocker_center_pt = Geom::Point3d.new(mid_pt.x, @front_well_pt2.y, 0)
            mid_pt = Geom::Point3d.new(@front_well_pt2.x+(@rocker_center_pt.x-@front_well_pt2.x)*(side_low_v_hash['mid_pt'].x-@well_pt.x)/(@side_low_pt.x-@well_pt.x), @front_well_pt2.y, 0)
            
            rocker_v_hash = Hash.new
            rocker_v_hash['mid_pt'] = mid_pt
            
            ux = rocker_u_hash['mid_pt'].x - rocker_center.x
            uz = rocker_u_hash['mid_pt'].z - rocker_center.z
            vx = side_high_center.x - side_high_v_hash['mid_pt'].x
            vz = side_high_center.z - side_high_v_hash['mid_pt'].z
            ax = side_high_v_hash['mid_pt'].x - rocker_u_hash['mid_pt'].x
            az = side_high_v_hash['mid_pt'].z - rocker_u_hash['mid_pt'].z
            phi = (ax*vz-az*vx)/(ux*vz-uz*vx)
            rocker_end_center = rocker_u_hash['mid_pt'] + (rocker_u_hash['mid_pt'] - rocker_center).transform!(Geom::Transformation.new(phi))
            
            @rocker_end_pt = Geom::Point3d.new((@rear_well_pt2.x+@rocker_center_pt.x)*0.5, @rocker_center_pt.y, 0)
            mid_pt = Geom::Point3d.new(@rocker_center_pt.x+(@rocker_end_pt.x-@rocker_center_pt.x)*(side_high_v_hash['mid_pt'].x-@side_low_pt.x)/(@side_high_pt.x-@side_low_pt.x), @rocker_center_pt.y, 0)
            rocker_end_v_hash = Hash.new
            rocker_end_v_hash['mid_pt'] = mid_pt
            
            temp1 = (bw_v_hash['mid_pt'] - side_high_center).normalize
            temp2 = (bw_v_hash['mid_pt'] - bw_center).normalize
            temp = Geom::Vector3d.new(temp2.x, temp2.y, temp2.x*temp1.z/temp1.x*1.1)
            temp_vec = temp*temp1
            mid_pt = Geom.intersect_line_plane([@roof_pt, (@roof_pt - @roof_end_pt)],[side_high_center, temp_vec])
            side_high_u_hash = Hash.new
            side_high_u_hash['mid_pt'] = mid_pt
            
            ux = side_high_u_hash['mid_pt'].x - side_high_center.x
            uz = side_high_u_hash['mid_pt'].z - side_high_center.z
            vx = temp2.x
            vz = temp2.z
            ax = bw_v_hash['mid_pt'].x + temp2.x - side_high_u_hash['mid_pt'].x
            az = bw_v_hash['mid_pt'].z + temp2.z - side_high_u_hash['mid_pt'].z
            phi = (ax*vz-az*vx)/(ux*vz-uz*vx)
            rear_well_center = side_high_u_hash['mid_pt'] + (side_high_u_hash['mid_pt'] - side_high_center).transform!(Geom::Transformation.new(phi))
            
            rear_well_u_hash = Hash.new
            rear_well_u_hash['mid_pt'] = @bw_pt
            
            mid_pt = @side_high_pt + (@side_high_pt - side_high_u_hash['mid_pt']).normalize.transform!(Geom::Transformation.new(@belt_cur_n))
            
            rocker_end_u_hash = Hash.new
            rocker_end_u_hash['mid_pt'] = mid_pt
            
            rear_well_v_hash = Wireframe.link(@side_high_pt, mid_pt3, @rear_well_pt, @numseg)
            
            rear_well_v2_hash = Wireframe.link_tan(@rocker_end_pt, @rear_well_pt2, @rocker_end_pt-rocker_end_v_hash['mid_pt'], (@rear_well_pt2.x-@rocker_end_pt.x)*0.5, @numseg)
            
            k = (@side_high_pt.x - rear_well_v_hash['mid_pt'].x)/(@side_high_pt.x - @rear_well_pt.x)
            rear_well_center2 = rocker_end_u_hash['mid_pt'] + (rear_well_u2_hash['mid_pt']-rocker_end_u_hash['mid_pt']).transform!(k)
            
            data['ctrlPts'] = [@hood_u_pt, well_u_hash['mid_pt'], @well_pt, fw_v_hash['mid_pt'], side_low_center, side_low_v_hash['mid_pt'], @fw_u_pt, side_low_u_hash['mid_pt'], @side_low_pt]
            side_low_surf = Bezier_surf.new(data)
            side_low_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@well_pt, front_well_u2_hash['mid_pt'], @front_well_pt2, side_low_v_hash['mid_pt'], rocker_center, rocker_v_hash['mid_pt'], @side_low_pt, rocker_u_hash['mid_pt'], @rocker_center_pt]
            rocker_surf = Bezier_surf.new(data)
            rocker_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@fw_u_pt, side_low_u_hash['mid_pt'], @side_low_pt, roof_v_hash['mid_pt'], side_high_center, side_high_v_hash['mid_pt'], @roof_pt, side_high_u_hash['mid_pt'], @side_high_pt]
            side_high_surf = Bezier_surf.new(data)
            side_high_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@side_low_pt, rocker_u_hash['mid_pt'], @rocker_center_pt, side_high_v_hash['mid_pt'], rocker_end_center, rocker_end_v_hash['mid_pt'], @side_high_pt, rocker_end_u_hash['mid_pt'], @rocker_end_pt]
            rocker_end_surf = Bezier_surf.new(data)
            rocker_end_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@roof_pt, side_high_u_hash['mid_pt'], @side_high_pt, bw_v_hash['mid_pt'], rear_well_center, rear_well_v_hash['mid_pt'], @bw_pt, rear_well_u_hash['mid_pt'], @rear_well_pt]
            rear_well_surf = Bezier_surf.new(data)
            rear_well_surf.create_patch(c3.entities)
            
            data['ctrlPts'] = [@side_high_pt, rocker_end_u_hash['mid_pt'], @rocker_end_pt, rear_well_v_hash['mid_pt'], rear_well_center2, rear_well_v2_hash['mid_pt'], @rear_well_pt, rear_well_u2_hash['mid_pt'], @rear_well_pt2]
            rear_well_surf2 = Bezier_surf.new(data)
            rear_well_surf2.create_patch(c3.entities)
            
#--------------------------------------------------------------------------------------------------------------------------------------------------------------
    # Component 5: Back
    # 2 variables
            
            c5 = aCar.entities.add_group
            @hip_n = 200.0
            @hip_mid_pt = Geom::Point3d.new((@box_r+@wheelbase+(1+@wheel_scale)*@r_wheel*@well_scale+@head_u1_pt.x)*0.5 + (@box_r-(@wheelbase+(1+@wheel_scale)*@r_wheel*@well_scale+@head_u1_pt.x))*0.5*rand(), 0, 0)
            hip_vec = (@bw_end_pt - profile1['bw_ctrl_pt']).normalize
            @hip_n = [(@hip_mid_pt.x-@bw_end_pt.x)/hip_vec.x, -@bw_end_pt.z/hip_vec.z].min*rand()
            
            temp = Wireframe.link_tan(@bw_end_pt, @hip_mid_pt, hip_vec, @hip_n, @numseg)
            profile1['hip'] = temp['all_pt']
            profile1['hip_ctrl_pt'] = temp['mid_pt']
            
            hip_center = Geom.intersect_line_plane([profile1['hip_ctrl_pt'], ey], [bw_center, (bw_center-roof_center)*(roof_center-fw_center)])
            
            hip_temp = Wireframe.link_tan(@hip_mid_pt, @rear_pt, ey, @hip_n, @numseg*2)
            
            @rear_start_pt = hip_temp['all_pt'][@numseg]
            temp_vec = @rear_pt - @hip_mid_pt
            mid_pt1 = Geom.intersect_line_line([@hip_mid_pt, ey], [@rear_start_pt, temp_vec])
            mid_pt2 = Geom.intersect_line_line([@rear_pt, @rear_pt - hip_temp['mid_pt']], [@rear_start_pt, temp_vec])
            rear_hash = Wireframe.link(@hip_mid_pt, mid_pt1, @rear_start_pt, @numseg)
            rear_u_hash = Wireframe.link(@rear_start_pt, mid_pt2, @rear_pt, @numseg)
            
            hip_u_vec = (@bw_pt - bw_v_hash['mid_pt']).normalize
            mid_pt = Geom.intersect_line_plane([@bw_pt, @bw_pt - bw_v_hash['mid_pt']], [hip_center, hip_u_vec])
            hip_v_hash = Hash.new
            hip_v_hash['mid_pt'] = mid_pt
            rear_center = Geom.intersect_line_plane([rear_well_u_hash['mid_pt'], rear_well_u_hash['mid_pt'] - rear_well_center], [hip_center, hip_u_vec])
            
            data['ctrlPts'] = [@bw_end_pt, bw_u_hash['mid_pt'], @bw_pt, profile1['hip_ctrl_pt'], hip_center, hip_v_hash['mid_pt'], @hip_mid_pt, rear_hash['mid_pt'], @rear_start_pt]
            hip_surf = Bezier_surf.new(data)
            hip_surf.create_patch(c5.entities)
            
            data['ctrlPts'] = [@bw_pt, rear_well_u_hash['mid_pt'], @rear_well_pt, hip_v_hash['mid_pt'], rear_center, rear_v_hash['mid_pt'], @rear_start_pt, rear_u_hash['mid_pt'], @rear_pt]
            rear_surf = Bezier_surf.new(data)
            rear_surf.create_patch(c5.entities)
            
            @rwheel_x = (@rear_pt.x + @rear_well_pt2.x)*0.5
            @rwheel_z = (@rear_pt.z + @rear_well_pt2.z)*0.5
            # Rear wheel extension
            rw1 = Geom::Point3d.new(@rwheel_x + @r_wheel*1.2, @box_w, 0.0)
            rw2 = Geom::Point3d.new(@rwheel_x, @box_w, @r_wheel*1.2)
            rw3 = Geom::Point3d.new(@rwheel_x - @r_wheel*1.2, @box_w, 0.0)
            rw12 = Wireframe.link_tan(rw2, rw1, Geom::Vector3d.new(1, 0, 0), 70, @numseg)
            rw23 = Wireframe.link_cur(rw1, rw12['mid_pt'], rw2, rw3, 1, @numseg)
            rw_mid2 = Geom.intersect_line_plane([@rear_well_pt, rw2 - @rear_well_pt], [Geom::Point3d.new(0.0, (@box_w*0.2+@rear_pt.y*0.8), 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            rw_mid1 = Geom.intersect_line_plane([@rear_pt, @rear_pt - rear_u_hash['mid_pt']], [Geom::Point3d.new(0.0, (@box_w*0.2+@rear_pt.y*0.8), 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            rw_mid3 = @rear_well_pt2 + (rw_mid1-@rear_pt).transform!(Geom::Transformation.scaling(-1, 1, 1))
            
            rw_center1 = Geom.intersect_line_plane([rear_v_hash['mid_pt'], rear_v_hash['mid_pt']-rear_center], [Geom::Point3d.new(0.0, (@box_w+@rear_pt.y)*0.5, 0.0), Geom::Vector3d.new(0.0, -1.0, 0.0)])
            rw_center2_hash = Wireframe.link_cur(rw_mid1, rw_center1, rw_mid2, rw_mid3, 1, @numseg)
            
            c5_wheel = aCar.entities.add_group
            data['ctrlPts'] = [@rear_well_pt, rw_mid2, rw2, rear_v_hash['mid_pt'], rw_center1, rw12['mid_pt'], @rear_pt, rw_mid1, rw1]
            rear_well_ext1_surf = Bezier_surf.new(data)
            rear_well_ext1_surf.create_patch(c5_wheel.entities)
            
            data['ctrlPts'] = [@rear_well_pt2, rw_mid3, rw3, rear_well_u2_hash['mid_pt'], rw_center2_hash['mid_pt'], rw23['mid_pt'], @rear_well_pt, rw_mid2, rw2]
            rear_well_ext2_surf = Bezier_surf.new(data)
            rear_well_ext2_surf.create_patch(c5_wheel.entities)
#===================================================================== Modeling ===============================================================================

#===================================================================== Post - Modeling ===============================================================================
            #-----------------------------------------------------------------------------
            #Wireframe.softenCoplanarGrp(aCar, 100.0)
            #-----------------------------------------------------------------------------
            #lift_z = Geom::Transformation.new(Geom::Point3d.new(0, 0, @clearance))
            #aCar.transform!(lift_z)
            scale = Geom::Transformation.new(@dimension_coef)
            aCar.transform!(scale)
            t_mirror = Geom::Transformation.scaling(1, -1, 1)
            aCar_mirror = aCar.copy#
            aCar_mirror.transform!(t_mirror)
#===================================================================== Post - Modeling ===============================================================================
    end #end main()



end ##end of class
